# pylint: disable=locally-disabled, multiple-statements, no-value-for-parameter, no-member, unused-argument, invalid-name, too-many-function-args, unused-import, missing-function-docstring
"""
TBD
"""
from timeit import default_timer as timer
from numba import vectorize
import numpy as np
from entangle.logging.debug import logging
from entangle.process import process


@process(timeout=0.5, shared_memory=True)
def dopow(names, smm=None, sm=None):
    print("Names:", names)
    (namea, nameb, _, _, typea, typeb) = names

    start = timer()

    # Get named shared memory segments
    shma = sm(namea)
    shmb = sm(nameb)

    # Get prepopulated array buffers
    np_shma = np.frombuffer(shma.buf, dtype=typea)
    np_shmb = np.frombuffer(shmb.buf, dtype=typeb)

    @vectorize(['float32(float32, float32)'], target='cuda')
    def _pow(a, b):
        return a ** b

    _pow(np_shma, np_shmb)

    duration = timer() - start

    print("Powers Time: ", duration)


@process(timeout=0.5, shared_memory=True)
def createvectors(smm=None, sm=None):

    vec_size = 100000000

    start = timer()

    # Create random array of values
    a = b = np.array(np.random.sample(vec_size), dtype=np.float32)

    # write matrices to shared memory
    shma = smm.SharedMemory(a.nbytes)
    shmb = smm.SharedMemory(b.nbytes)

    names = (shma.name, shmb.name, a.shape, b.shape, a.dtype, b.dtype)

    duration = timer() - start
    print("Create Vectors Time: ", duration)

    return names


dp = dopow(
    createvectors()
)

dp()



"""
__version__.py
~~~~~~~~~~~~~~

Information about the current version of the py-package-template package.
"""

__title__ = 'py-entangle'
__description__ = 'A python native parallel processing framework based on simple decorators.'
__version__ = 'VERSION'
__author__ = 'Darren Govoni'
__author_email__ = 'darren@ontrenet.com'
__license__ = 'MIT'
__url__ = 'https://github.com/radiantone/entangle'



"""
docker.py - Module that provides docker support decorators for running tasks inside containers
"""
import logging
import inspect
import re
from functools import partial
import docker as dkr


client = dkr.from_env()


def docker(function=None,
           image=None,
           packages=None,
           consume_gpu=True):
    """

    :param function:
    :param image:
    :param sleep:
    :return:
    """
    def decorator(func):
        def wrapper(f_func):

            lines = inspect.getsource(f_func)
            logging.info("Running container: %s",image)
            lines = re.sub('@', '#@', lines)
            name = f_func.__name__

            installpackages = []
            for package in packages:
                installpackages += ["pip install -q {}".format(package)]

            code = "bash -c \"{}\npython <<HEREDOC {}\nprint({}())\"\nHEREDOC".format(
                ";".join(installpackages), lines, name)

            logging.debug(code)
            env = {}

            if not consume_gpu:
                env['TF_FORCE_GPU_ALLOW_GROWTH'] = True

            result = client.containers.run(
                image, code, environment=env)

            logging.debug(result)
            return result

        pfunc = partial(wrapper, func)
        pfunc.__name__ = func.__name__

        return pfunc

    if function is not None:
        return decorator(function)

    return decorator



# pylint: disable=locally-disabled, multiple-statements, too-few-public-methods, unused-argument, missing-function-docstring
"""
importer.py - Automatically import missing modules from pypi
"""
from importlib import util
import subprocess
import sys


class PipFinder:
    """
    Automatically import missing modules from pypi
    """
    @classmethod
    def find_spec(cls, name, path, target=None):
        print(f"Module {name!r} not installed.  Attempting to pip install")
        cmd = f"{sys.executable} -m pip install {name}"
        try:
            subprocess.run(cmd.split(), check=True)
        except subprocess.CalledProcessError:
            return None

        return util.find_spec(name)


sys.meta_path.append(PipFinder)



# pylint: disable=locally-disabled, multiple-statements, no-value-for-parameter, invalid-name, too-many-function-args, unused-import, missing-function-docstring
"""
TBD
"""
from entangle.logging.debug import logging
from entangle.process import process
from multiprocessing import Process


@process
def one():
    return 1


@process
def two():
    return 2


@process
def five():
    return 5


@process
def num(n):
    return n


@process
def add(a, b):
    v = int(a) + int(b)
    print("ADD: *"+str(v)+"*")
    return v


@process
def subtract(a, b):
    return int(a) - int(b)


if __name__ == '__main__':

    """
    _five = five()
    _two = two()
    _sub = subtract(_five,_two)
    _num = num(6)
    _two2 = two() if False else one()
    _add1 = add(_num,_two2)
    result = add(_add1,_sub)
    """

    workflow = add(
        add(
            num(6),
            two() if False else one()
        ),
        subtract(
            five(),
            add(
                subtract(
                    num(8),
                    two()
                ),
                one()
            )
        )
    )

    def callback(result):
        print("CALLBACK:", result.result())

    # set up future callbacks
    future = workflow.future(callback=callback)
    print('Future:', future)
    # Trigger workflow. Does not block
    workflow(proc=True)

    # Notify all the futures
    future.entangle() # Does not block



# pylint: disable=locally-disabled, multiple-statements, line-too-long, no-value-for-parameter, invalid-name, too-many-function-args, unused-import, missing-function-docstring
"""
sshschedulerexample.py - Example usage of @ssh and @schedular decorators

Create and copy ssh keys to your hosts

$ ssh-keygen -t rsa
$ ssh-copy-id -i /home/$USER/.ssh/id_rsa.pub <destination hostname>

This must be done bidirectionally between all hosts

"""
from entangle.logging.debug import logging
from entangle.process import process
from entangle.thread import thread
from entangle.ssh import ssh
from entangle.workflow import workflow
from entangle.scheduler import scheduler


scheduler_config = {'cpus': 3,
                    'impl': 'entangle.scheduler.DefaultScheduler'}


class MyResult:
    """
    MyResult
    """

    def __init__(self):
        self._result = None

    def set_result(self, value):
        self._result = value

    def get_result(self):
        return self._result


@scheduler(**scheduler_config)
@process
def add(a, b):
    v = int(a.get_result()) + int(b)
    logging.info("ADD: *"+str(v)+"*")
    return v


@ssh(user='darren', host='miko', key='/home/darren/.ssh/id_rsa.pub', env="venv", python='/home/darren/venv/bin/python')
@scheduler(**scheduler_config)
@process
def two():
    logging.info("Returning 2")
    return 2


@ssh(user='darren', host='phoenix', key='/home/darren/.ssh/id_rsa.pub', python='/home/darren/miniconda3/bin/python')
@scheduler(**scheduler_config)
@thread
def three():
    logging.info("Returning 3")
    _result = MyResult()
    _result.set_result(3)
    return _result


@ssh(user='darren', host='radiant', key='/home/darren/.ssh/id_rsa.pub', python='/home/darren/venv/bin/python')
@scheduler(**scheduler_config)
@process
def workflow2():

    _add = add(
        three(),
        two()
    )

    _result = _add()
    my_result = MyResult()
    my_result.set_result(_result)
    return my_result


if __name__ == '__main__':
    workflow = workflow2()
    result = workflow()
    print("WORKFLOW OBJ:", result)
    print("WORKFLOW RESULT:", result.get_result())



"""

tests.conftest.py
~~~~~~~~~~~~~~~~~

This module is used by PyTest to detect test fixtures that are used by
multiple test modules. For more imformation on fixtures in PyTest, see
https://docs.pytest.org/en/latest/fixture.html.
"""

from typing import Any, Dict, List

import pytest



# pylint: disable=locally-disabled, multiple-statements, no-value-for-parameter, invalid-name, too-many-function-args, unused-import, missing-function-docstring
"""
TBD
"""
import threading
import time
from entangle.logging.info import logging
from entangle.dataflow import thread
from entangle.dataflow import process
from entangle.dataflow import dataflow


def triggered(func, result):
    print("triggered: {} {}".format(func.__name__, result))


@dataflow(callback=triggered)
@thread
def printx(x):
    print('   printx: {} {}'.format(x, threading.current_thread().name))
    return "X: {}".format(x)


@dataflow(callback=triggered)
@thread
def printy(y):
    print('printy: {}'.format(threading.current_thread().name))
    return "Y: {}".format(y)


@dataflow(callback=triggered)
@thread
def printz(z):
    print('printz: {}'.format(threading.current_thread().name))
    return "Z: {}".format(z)


@dataflow(callback=triggered, maxworkers=3)
def emit(a):
    print('emit: {}'.format(threading.current_thread().name))
    return a+"!"


# Create the dataflow graph
# Defer whether we will forward to printx() or printy() depending
# on the result receive from emit. This won't be known until the data is ready.
flow = emit(
    lambda x: printx(
        printz()
    )
    if x == 'emit' else
    printy()
)

# Invoke the dataflow graph with initial input
flow('emit')

time.sleep(2)
print("----------------------------")
# Call flow again with different input value
flow('HELLO')



"""
tests.test_entangle.py
~~~~~~~~~~~~~~~~~~~~

"""


class TestEntangle:
    """Test suite for entangle app class."""

    def test_dataflowexample2(self):
        from entangle.logging.debug import logging
        from entangle.examples.dataflowexample2 import emit, printx, printy, printz, echo

        flow = emit(
            printx(
                printz(
                    echo()
                )
            ),
            printy(
                printz()
            ),
            printy()
        )
        try:
            flow('emit')
            assert True
        except:
            assert False


    def test_dataflowexample(self):
        from entangle.logging.debug import logging
        from entangle.examples.dataflowexample import emit, printx, printy, printz, echo


        flow = emit(
            printx(
                printz(
                    echo()
                )
            ),
            printy(
                printz()
            ),
            printy()
        )
        try:
            flow('emit')
            assert True
        except:
            assert False


    def test_schedulerexample2(self):
        from entangle.logging.info import logging
        from entangle.examples.schedulerexample2 import workflow2

        result = workflow2()
        try:
            result()
            assert True
        except:
            assert False


    def test_schedulerexample(self):
        from entangle.logging.info import logging
        from entangle.examples.schedulerexample import workflow2

        result = workflow2()
        try:
            result()
            assert True
        except:
            assert False


    def test_listexample2(self):
        from entangle.logging.info import logging
        from entangle.examples.listexample2 import emit, printx, printy, printz

        flow = emit(
            lambda x: list(map(lambda f: f(x), [
                lambda y: [printx(
                    printz()
                ) for _ in range(0, 10)],
                printy(
                    printz()
                )
            ]))[0]
        )

        try:
            flow('emit')
            assert True
        except:
            assert False


    def test_listexample(self):
        from entangle.logging.info import logging
        from entangle.examples.listexample import emit, printx, printy, printz

        flow = emit(
            lambda x: [printy(
                printz(
                    printx()
                )
            ) for _ in range(0, 10)]
        )

        try:
            flow('emit')
            assert True
        except:
            assert False

    def test_lambdaexample(self):
        from entangle.logging.info import logging
        from entangle.examples.lambdaexample import emit, printx, printy, printz

        flow = emit(
            lambda x: printx(
                printz()
            )
            if x == 'emit' else
            printy()
        )

        try:
            flow('emit')
            assert True
        except:
            assert False

    def test_example6(self):
        from entangle.logging.info import logging
        from entangle.examples.example6 import workflow1, workflow2

        result = workflow2(workflow1)

        try:
            result()
            assert True
        except:
            assert False

    def test_example5(self):
        from entangle.logging.info import logging
        from entangle.examples.example5 import workflow1, workflow2

        result = workflow2(workflow1)

        try:
            result()
            assert True
        except:
            import traceback
            print(traceback.format_exc())
            assert False

    def test_example2(self):
        from entangle.logging.debug import logging
        from entangle.examples.example2 import workflow1, workflow2

        result = workflow2(workflow1)

        try:
            result()
            assert True
        except:
            assert False

    def test_example(self):
        from entangle.logging.debug import logging
        from entangle.examples.example import add, num, one, two, five, subtract

        result = add(
            add(
                num(6),
                two() if False else one()
            ),
            subtract(
                five(),
                two()
            )
        )
        assert result() == 10



# pylint: disable=locally-disabled, multiple-statements, import-outside-toplevel, import-error, no-value-for-parameter, invalid-name, too-many-function-args, unused-import, missing-function-docstring
"""
aiexample.py
"""
from entangle.logging.debug import logging
from entangle.containers import docker
from entangle.process import process


@process()
@docker(image="tensorflow/tensorflow:latest-gpu", packages=['tensorflow_datasets'])
def train():
    import tensorflow as tf
    import tensorflow_datasets as tfds

    (ds_train, ds_test), ds_info = tfds.load(
        'mnist',
        split=['train', 'test'],
        shuffle_files=True,
        as_supervised=True,
        with_info=True,
    )

    def normalize_img(image, label):
        return tf.cast(image, tf.float32) / 255., label

    ds_train = ds_train.map(
        normalize_img, num_parallel_calls=tf.data.experimental.AUTOTUNE)
    ds_train = ds_train.cache()
    ds_train = ds_train.shuffle(ds_info.splits['train'].num_examples)
    ds_train = ds_train.batch(128)
    ds_train = ds_train.prefetch(tf.data.experimental.AUTOTUNE)

    ds_test = ds_test.map(
        normalize_img, num_parallel_calls=tf.data.experimental.AUTOTUNE)
    ds_test = ds_test.batch(128)
    ds_test = ds_test.cache()
    ds_test = ds_test.prefetch(tf.data.experimental.AUTOTUNE)

    model = tf.keras.models.Sequential([
        tf.keras.layers.Flatten(input_shape=(28, 28)),
        tf.keras.layers.Dense(128, activation='relu'),
        tf.keras.layers.Dense(10)
    ])

    model.compile(
        optimizer=tf.keras.optimizers.Adam(0.001),
        loss=tf.keras.losses.SparseCategoricalCrossentropy(from_logits=True),
        metrics=[tf.keras.metrics.SparseCategoricalAccuracy()],
    )

    model.fit(
        ds_train,
        epochs=6,
        verbose=0,
        validation_data=ds_test,
    )

    return model.summary()


result = train()
print(result())



# pylint: disable=locally-disabled, not-callable, line-too-long, unused-argument, too-few-public-methods, no-self-use

"""
scheduler.py - A shared memory scheduler that runs in a process within a workflow. nodes send requests to the scheduler queue
and then wait for a reply. the reply gives the node parameters to run its process such as what cpu to run it on. When a process is finished,
the node sends message to scheduler that its processed finished and the scheduler can then task another node.

"""

from importlib import import_module
from multiprocessing import Queue
from functools import partial
import logging
import threading
import inspect
import os
import types
from typing import Callable
from entangle.process import ProcessMonitor
from entangle.thread import ThreadMonitor


CPUS = []

queue = Queue()

CMD = "/usr/bin/lscpu -p=socket,cpu,online"
stream = os.popen(CMD)
output = stream.readlines()
for line in output:
    if line[0] == '#':
        continue
    _cpu = line.strip().split(',')
    if _cpu[2] == 'Y':
        CPUS += [_cpu]
        # Put CPU cookie on queue
        queue.put(_cpu)

logging.debug('CPUS: %s',CPUS)


def import_string(dotted_path):
    """
    Import a dotted module path and return the attribute/class designated by the
    last name in the path. Raise ImportError if the import failed.
    """

    module_path, class_name = dotted_path.rsplit('.', 1)

    module = import_module(module_path)

    return getattr(module, class_name)


class FileLockScheduler:
    """
    Will implement de-centralized CPU binding using shared memory and lock files
    see: entangle/scratch/sheduler.py
    """


class DefaultScheduler:
    """
    Desc
    """
    def register(self, f_func, cpus=12):
        """
        Desc
        :param f:
        :param cpus:
        :return:
        """
        def schedule(_func, *args, **kwargs):
            """
            Desc
            :param f:
            :param args:
            :param kwargs:
            :return:
            """
            logging.debug("DefaultScheduler: args %s",str(args))
            logging.debug("DefaultScheduler: before:")

            logging.debug("DefaultScheduler: thread %s",
                threading.current_thread().name)

            logging.debug("Waiting on CPU")
            cpu_pending = True

            while cpu_pending:
                cpu = queue.get()
                if int(cpu[1]) >= int(cpus):
                    logging.debug(
                        "     CPU not within allocation: %s %s",cpu, cpus)
                    queue.put(cpu)
                else:
                    logging.debug("GRABBED CPU: %s %s",cpu, cpus)
                    break

            logging.debug("GOT CPU: %s",cpu)
            logging.debug(_func)

            if not isinstance(_func, types.FunctionType):
                kwargs['cpu'] = cpu[1]
                kwargs['scheduler'] = queue

            if cpu:
                pid = os.getpid()
                cpu_mask = [int(cpu[1])]
                logging.debug("Setting cpu_mask %s",cpu_mask)
                os.sched_setaffinity(pid, cpu_mask)

            result = _func(*args, **kwargs)
            logging.debug("Putting cpu %s back on scheduler queue",cpu)

            queue.put(cpu)
            logging.debug("DefaultScheduler: after")
            logging.debug("DefaultScheduler: return %s",result)
            return result

        return partial(schedule, f_func)


def scheduler(function=None,
              impl='entangle.scheduler.DefaultScheduler',
              cpus=12,
              algorithm='first_available',
              max_time=60*60) -> Callable:
    """
    Desc
    :param function:
    :param impl:
    :param cpus:
    :param algorithm:
    :param max_time:
    :return:
    """

    _scheduler = import_string(impl)()

    logging.debug("scheduler: Requesting %s cpus",cpus)

    def decorator(func, cpus=12) -> Callable:
        """
        Desc
        :param func:
        :param cpus:
        :return:
        """

        _func = func

        if isinstance(func, (ProcessMonitor, ThreadMonitor)):
            _func = func.func

        if isinstance(_func,partial):

            def find_func(pfunc):
                if isinstance(pfunc, partial):
                    return find_func(pfunc.func)
                return pfunc

            _func = find_func(_func)

        source = inspect.getsource(_func)
        logging.debug("scheduler: source: %s",source)

        def wrapper(_wfunc, *args, **kwargs) -> Callable:
            logging.debug("scheduler: Calling function: %s", str(_wfunc))
            logging.debug("Waiting 2 seconds...")
            # time.sleep(2)
            return _wfunc(*args, **kwargs)

        logging.debug("scheduler: decorator %s cpus",cpus)
        logging.debug("scheduler: Registering function: %s",str(func))
        sfunc = _scheduler.register(func, cpus=cpus)
        logging.debug("scheduler: Returning function: %s",str(sfunc))
        _pfunc = partial(wrapper, sfunc)
        _pfunc.source = source

        if isinstance(func, (ProcessMonitor, ThreadMonitor)):
            _pfunc.__name__ = func.func.__name__
        else:
            _pfunc.__name__ = func.__name__

        return _pfunc


    if function is not None:
        _decorator = decorator(function, cpus=cpus)
        return _decorator

    pfunc = partial(decorator, **{'cpus': cpus})
    logging.debug("scheduler: no source: ")

    return pfunc



"""
workflow.py - Module that provides workflow decorator
"""
import logging

from functools import partial
from concurrent.futures import ThreadPoolExecutor
from concurrent.futures import ProcessPoolExecutor

global PROCESSPOOL, THREADPOOL
PROCESSPOOL = None
THREADPOOL = None   


def process(func):
    """
    Desc
    :param func:
    :param kwargs:
    :return:
    """
    def inner(*args, **kwargs):
        """
        Desc
        :param args:
        :param kwargs:
        :return:
        """
        _workflow = func(*args, **kwargs)
        return _workflow

    inner.thread = False
    inner.func = func
    return inner


def thread(func):
    """
    Desc
    :param func:
    :param kwargs:
    :return:
    """
    def inner(*args, **kwargs):
        """
        Desc
        :param args:
        :param kwargs:
        :return:
        """
        _workflow = func(*args, **kwargs)

        return _workflow

    inner.thread = True
    inner.func = func
    return inner


class DataflowNode:
    """
    Desc
    """

    def __init__(self, func, *args, **kwargs):
        """
        Desc
        :param func:
        :param args:
        :param kwargs:
        """
        self.func = func
        self.args = args

        self.thread = func.thread if hasattr(func, 'thread') else True
        self.partial = partial(func, *self.args)
        self.partial.__name__ = func.__name__

        self.result = kwargs['result'] if 'result' in kwargs else []
        self.callback = kwargs['callback'] if 'callback' in kwargs else None

        logging.debug(self.partial)
        logging.debug("%s self.args", self.args)

        # Build dataflow DAG here

    def __call__(self, *args, **kwargs):

        logging.debug("Inside dataflow: %s",
            self.func.__name__)

        if len(self.args) > 0 and not isinstance(self.args[0], DataflowNode):
            logging.debug("Calling partial")

            if callable(self.args[0]) and self.args[0].__name__ == "<lambda>":
                result = self.args[0](*args)
                if isinstance(result, list):
                    for _r in result:
                        _r.__name__ = 'lambda'
                        _df = DataflowNode(_r,*args)
                        _df(*args)
                    return self
                else:
                    result(*args)
                    return self
            else:
                result = self.partial()

            logging.debug("Result %s",result)

            if self.callback:
                logging.debug('Calling callback %s',self.partial.__name__)
                THREADPOOL.submit(self.callback, self.partial.func, result)

            for arg in args:
                if isinstance(arg, DataflowNode):
                    if isinstance(arg, partial):
                        logging.debug("Sending %s to %s",
                            result, arg.func.__name__)

            for arg in args:

                if hasattr(arg,'func'):
                    logging.debug('   LAUNCHING: %s %s',
                        arg.func.__name__, result)
                    logging.debug('     Thread: %s',arg.thread)
                    logging.debug("Calling %s('%s')",
                        arg.func.__name__, result)

                if hasattr(arg,'thread') and arg.thread:
                    THREADPOOL.submit(arg, result, **{'result': self.result})
                else:
                    PROCESSPOOL.submit(arg, result, **{'result': self.result})
        else:
            logging.debug("Calling with args %s %s",
                self.func.__name__, *args)

            result = self.func(*args)

            if self.callback:
                logging.debug('Calling callback %s',self.func.__name__)
                THREADPOOL.submit(self.callback, self.func, result)

            logging.debug("Result %s",result)
            for arg in self.args:
                if isinstance(arg, DataflowNode):
                    logging.debug("Sending %s to %s",
                        result, arg.func.__name__)

            for arg in self.args:
                logging.debug('   LAUNCHING: %s %s',
                    arg.func.__name__, result)
                logging.debug('     Thread: %s',arg.thread)
                logging.debug("Calling %s('%s')",
                    arg.func.__name__, result)

                if arg.thread:
                    THREADPOOL.submit(arg, result)
                else:
                    PROCESSPOOL.submit(arg, result)

        return self


def dataflow(function=None,
             callback=None,
             maxworkers=3):
    """
    Desc
    :param function:
    :param callback:
    :param maxworkers:
    :return:
    """
    global PROCESSPOOL, THREADPOOL

    def decorator(func):
        def wrapper(f_func, *dargs, **dkwargs):
            logging.debug("Calling decorator %s", f_func.__name__)
            logging.debug("dargs %s",str(dargs))

            def invoke(*args, **kwargs):
                logging.debug("invoke f %s %s", f_func, args)
                kwargs['callback'] = callback
                return DataflowNode(f_func, *args, **kwargs)

            return invoke

        return wrapper(func)

    try:
        PROCESSPOOL = ProcessPoolExecutor(max_workers=maxworkers)
        THREADPOOL = ThreadPoolExecutor(max_workers=maxworkers)

        if function is not None:
            return decorator(function)

        return decorator
    finally:
        pass



# pylint: disable=locally-disabled, multiple-statements, no-value-for-parameter, invalid-name, too-many-function-args, unused-import, missing-function-docstring
"""
TBD
"""
import threading
import time
from entangle.logging.debug import logging
from entangle.dataflow import thread
from entangle.dataflow import process
from entangle.dataflow import dataflow


def triggered(func, result):
    print("triggered: {} {}".format(func.__name__, result))


@dataflow(callback=triggered)
@thread
def printx(x):
    print('printx: {}'.format(threading.current_thread().name))
    return "   X: {}".format(x)


@dataflow(callback=triggered)
@thread
def printy(y):
    print('printy: {}'.format(threading.current_thread().name))
    return "Y: {}".format(y)


@dataflow(callback=triggered)
@thread
def printz(z):
    print('printz: {}'.format(threading.current_thread().name))
    return "     Z: {}".format(z)


@dataflow(callback=triggered, maxworkers=3)
def emit(a):
    print('emit: {}'.format(threading.current_thread().name))
    return a+"!"


flow = emit(
    lambda x: list(map(lambda f: f(x), [
        lambda y: [printx(
            printz()
        ) for _ in range(0, 10)],
        printy(
            printz()
        )
    ]))[0]
)

flow('emit')


time.sleep(2)



# pylint: disable=locally-disabled, multiple-statements, no-value-for-parameter, invalid-name, too-many-function-args, unused-import, missing-function-docstring
"""
dataflowexamply.py
"""
import threading
import time

from entangle.logging.info import logging
from entangle.dataflow import thread
from entangle.dataflow import dataflow


def triggered(func, result):
    print("triggered: {} {}".format(func.__name__, result))


@dataflow(callback=triggered)
@thread
def printx(x):
    print('printx: {}'.format(threading.current_thread().name))
    return "X: {}".format(x)


@dataflow(callback=triggered)
@thread
def printy(y):
    print('printy: {}'.format(threading.current_thread().name))
    return "Y: {}".format(y)


@dataflow(callback=triggered)
@thread
def printz(z):
    print('printz: {}'.format(threading.current_thread().name))
    return "Z: {}".format(z)


@dataflow(callback=triggered)
@thread
def echo(e):
    print('echo: {}'.format(threading.current_thread().name))
    return "Echo! {}".format(e)


@dataflow(callback=triggered, maxworkers=3)
def emit(value):
    print('emit: {}'.format(threading.current_thread().name))
    return value+"!"


results = []

# Create the dataflow graph
flow = emit(
    printx(
        printz(
            echo()
        )
    ),
    printy(
        printz()
    ),
    printy()
)

# Invoke the dataflow graph with initial input
flow('emit')

time.sleep(2)

# Call flow again with different input value
flow('HELLO')



"""
tests.test_entangle.py
~~~~~~~~~~~~~~~~~~~~

"""
from entangle.process import process


class TestEntangle:
    """Test suite for entangle app class."""

    def test_example4(self):
        from entangle.examples.example4 import dopow, createvectors

        result = dopow(
            createvectors()
        )

        try:
            result()
            assert True
        except:
            assert False

    def test_example3(self):
        from entangle.examples.example3 import durations, dovectors1, dovectors2

        result = durations(
            dovectors1(),
            dovectors2()
        )

        try:
            result()
            assert True
        except:
            assert False



"""
process.py - Module that provides native OS process implementation of function tasks with support for shared memory
"""
import asyncio
import logging
import sys
import os
import inspect
import threading
import multiprocessing
import time
import json
import traceback
import queue as que
import signal

from typing import Callable
from functools import partial
from multiprocessing import Queue, Process
from multiprocessing.shared_memory import SharedMemory
from multiprocessing.managers import SharedMemoryManager

SMM = SharedMemoryManager()
SMM.start()

graph_queue = Queue()


def handler():
    """
    Handle any cleanup here
    """
    SMM.shutdown()


signal.signal(signal.SIGINT, handler)
signal.signal(signal.SIGTERM, handler)


def process(function=None,
            timeout=None,
            wait=None,
            cache=False,
            shared_memory=False,
            sleep=0) -> Callable:
    """

    :param function:
    :param timeout:
    :param cache:
    :param shared_memory:
    :param sleep:
    :return:
    """
    logging.debug("TIMEOUT: %s", timeout)

    def decorator(func) -> Callable:
        """
        Description
        :param func:
        :return:
        """
        def wrapper(f_func) -> Callable:
            """
            Description
            :param f_func:
            :return:
            """
            logging.debug(
                "ProcessMonitor: %s with wait %s", f_func, wait)
            return ProcessMonitor(f_func,
                                  timeout=timeout,
                                  wait=wait,
                                  shared_memory=shared_memory,
                                  cache=cache,
                                  sleep=sleep)

        return wrapper(func)

    if function is not None:
        return decorator(function)

    return decorator


class ProcessTerminatedException(Exception):
    """
    Description
    """


class ProcessTimeoutException(Exception):
    """
    Description
    """


class ProcessMonitor:
    """
    Primary monitor class for processes. Creates and monitors queues and processes to resolve argument tasks.
    """
    source = None
    graph = {}

    def __init__(self, func, *args, **kwargs) -> Callable:
        """

        :param func:
        :param args:
        :param kwargs:
        """
        self.func = func
        self.shared_memory = kwargs['shared_memory']
        self.sleep = kwargs['sleep']
        self.cache = kwargs['cache']
        self.timeout = kwargs['timeout']
        self.wait = kwargs['wait']
        self.execute = kwargs['execute'] if 'execute' in kwargs else True
        self.source = None
        self.result_queue = Queue()

    def get_func(self):
        """
        Desc
        """
        return self.func

    def future(self, callback=None):
        """
        Desc
        """
        asyncio.set_event_loop(asyncio.new_event_loop())
        loop = asyncio.get_event_loop()

        @asyncio.coroutine
        def wait_for_done(future_queue):

            logging.debug(
                "wait_for_done: looping: result queue: %s", future_queue)
            while True:
                try:
                    logging.debug("wait_for_done: checking queue")
                    result = future_queue.get_nowait()
                    logging.debug("wait_for_done: got result")
                    return result
                except:
                    yield

        _future = loop.create_task(wait_for_done(self.result_queue))

        def entangle():
            loop.run_until_complete(_future)

        _future.loop = loop
        _future.entangle = entangle

        if callback:
            _future.add_done_callback(callback)

        return _future

    def __call__(self, *args, **kwargs) -> Callable:
        """

        :param args:
        :param kwargs:
        :return:
        """

        logging.info("Process:invoke: %s", self.func.__name__)
        _func = self.func
        if isinstance(self.func, partial):

            def find_func(pfunc):
                if isinstance(pfunc, partial):
                    return find_func(pfunc.func)
                return pfunc

            _func = find_func(self.func)

        self.source = inspect.getsource(_func)

        def assign_cpu(func, cpu, **kwargs):
            """
            Desc
            :param func:
            :param cpu:
            :param kwargs:
            :return:
            """
            pid = os.getpid()
            cpu_mask = [int(cpu)]
            os.sched_setaffinity(pid, cpu_mask)

            func(**kwargs)

        def invoke(func, *args, **kwargs):
            """

            :param func:
            :param args:
            :param kwargs:
            :return:
            """
            graphs = []
            json_graph = "{}"
            json_graphs = []
            future_queue = kwargs['future_queue']
            del kwargs['future_queue']

            @asyncio.coroutine
            def get_result(_queue, func, sleep, now, process, event, wait, timeout):
                """

                :param q:
                :param func:
                :param sleep:
                :param now:
                :param process:
                :param timeout:
                :return:
                """

                if hasattr(func, '__name__'):
                    name = func.__name__
                else:
                    name = func

                logging.debug("get_result: started %s", now)
                while True:
                    logging.debug("Checking queue for result...")
                    try:
                        logging.debug(
                            "Waiting on event for %s with wait %s", name, self.wait)

                        if wait:
                            logging.debug(
                                "Wait event timeout in %s seconds.", wait)
                            event.wait(wait)
                            if not event.is_set():
                                if process.is_alive():
                                    process.terminate()
                                raise ProcessTimeoutException()
                        else:
                            logging.debug("Waiting until complete.")
                            event.wait()

                        logging.debug("Got event for %s", name)

                        logging.debug("Timeout is %s", timeout)
                        if timeout:
                            logging.debug(
                                "Pre get(timeout=%s)", timeout)
                            _response = _queue.get(timeout=timeout)
                            _result = _response['result']
                            logging.debug(
                                "Post get(timeout=%s)", timeout)
                        else:
                            _response = _queue.get()
                            _result = _response['result']

                        logging.debug("Got result for[%s] %s",
                                      name, str(_result))

                        yield

                        # Unwrap graph data list, and result (graph, result)

                        return _response
                    except multiprocessing.TimeoutError as ex:
                        logging.debug("Timeout exception")
                        raise ProcessTimeoutException() from ex
                    except que.Empty as ex:
                        if process and not process.is_alive():
                            raise ProcessTerminatedException() from ex

                        yield time.sleep(sleep)

            scheduler = None
            cpu = None

            if 'cpu' in kwargs:
                cpu = kwargs['cpu']
                del kwargs['cpu']

                if 'scheduler' in kwargs:
                    scheduler = kwargs['scheduler']
                    del kwargs['scheduler']

            if len(args) == 0:
                # Do nothing
                pass
            else:
                asyncio.set_event_loop(asyncio.new_event_loop())
                loop = asyncio.get_event_loop()

                _tasks = []
                processes = []

                for arg in args:

                    event = multiprocessing.Event()

                    if hasattr(arg, '__name__'):
                        aname = arg.__name__
                    else:
                        aname = arg

                    _queue = Queue()

                    _process = None

                    if isinstance(arg, partial):
                        logging.info("Process: %s", arg.__name__)

                        kargs = {'queue': _queue, 'event': event}
                        # If not shared memory

                        # if shared memory, set the handles
                        if self.shared_memory:
                            kargs['smm'] = SMM
                            kargs['sm'] = SharedMemory

                        if cpu:
                            arg_cpu = scheduler.get()

                            # TODO: Fix. This bypasses the scheduler logic of capping the CPU #'s.
                            logging.debug(
                                'ARG CPU SET TO: %s', arg_cpu[1])

                            # Update sharedlist with func to arg names
                            _process = Process(
                                target=assign_cpu, args=(
                                    arg, arg_cpu[1],), kwargs=kargs
                            )
                            _process.cookie = arg_cpu[1]
                        else:
                            logging.debug('NO CPU SET')

                            # Update sharedlist with func to arg names
                            _process = Process(
                                target=arg, kwargs=kargs)

                        if self.shared_memory:
                            _process.shared_memory = True

                        processes += [_process]

                        _process.start()
                    else:
                        logging.info("Value: %s", aname)

                        _queue.put(
                            {'graph': [(func.__name__, aname)], 'result': arg})
                        event.set()

                    now = time.time()

                    # Create an async task that monitors the queue for that arg
                    # It will wait for event set from this child process
                    if hasattr(arg, '__name__'):
                        graphs += [(func.__name__, arg.__name__)]
                    else:
                        graphs += [(func.__name__, arg)]
                    _tasks += [get_result(_queue, arg,
                                          self.sleep, now, _process, event, self.wait, self.timeout)]

                    # Wait until all the processes report results
                    tasks = asyncio.gather(*_tasks)

                # Ensure we have joined all spawned processes

                _args = loop.run_until_complete(tasks)
                try:
                    args = [_arg['result'] for _arg in _args]
                    arg_graph = [_arg['graph'] for _arg in _args]
                    json_graphs = [_arg['json']
                                   for _arg in _args if 'json' in _arg]
                except:
                    args = [_arg for _arg in _args]
                    arg_graph = []
                    json_graphs = []

                logging.debug("JSON GRAPHs: %s", json_graphs)
                logging.debug("ARG GRAPH: %s", arg_graph)

                def add_to_graph(gr, argr):
                    for item in argr:
                        if isinstance(item, list):
                            add_to_graph(gr, item)
                        elif isinstance(item, tuple):
                            gr += [item]

                    return gr

                logging.debug("GRAPH: %s", graphs)

                _G = {}
                _G[func.__name__] = {}
                G = _G[func.__name__]
                for node in graphs:
                    if len(node) < 2:
                        continue
                    if node[1] not in G:
                        G[node[1]] = []

                    for graphnode in json_graphs:
                        if node[1] in graphnode:
                            G[node[1]] = graphnode[node[1]]

                json_graph = json.dumps(_G, indent=4)
                logging.debug("JSON: %s", json_graph)
                _ = [process.join() for process in processes]

                # Put CPU cookie back on scheduler queue
                if scheduler:
                    for _process in processes:
                        logging.debug(
                            "Putting CPU: %s  back on scheduler queue.", _process.cookie)
                        scheduler.put(('0', _process.cookie, 'Y'))

            if cpu:
                pid = os.getpid()
                cpu_mask = [int(cpu)]
                os.sched_setaffinity(pid, cpu_mask)

            is_proc = False
            if 'proc' in kwargs and kwargs['proc'] is True:
                del kwargs['proc']
                is_proc = True

            event = None
            if 'event' in kwargs:
                event = kwargs['event']
                del kwargs['event']

            if 'queue' in kwargs:
                queue = kwargs['queue']
                # get the queue and delete the argument
                del kwargs['queue']

                # Pass in shared memory handles
                if self.shared_memory:
                    kwargs['smm'] = SMM
                    kwargs['sm'] = SharedMemory

                logging.info("Calling %s", func.__name__)
                logging.debug(args)

                if not cpu and 'cpu' in kwargs:
                    cpu = kwargs['cpu']
                    del kwargs['cpu']

                if not scheduler and 'scheduler' in kwargs:
                    scheduler = kwargs['scheduler']
                    del kwargs['scheduler']

                try:
                    if self.execute:
                        logging.debug("process: execute: %s", self.execute)

                        if is_proc:

                            logging.debug(
                                "self.execute: proc: creating process")
                            _mq = Queue()

                            def func_wrapper(_wf, _wq):
                                logging.debug("func_wrapper: %s", _wf)
                                result = _wf()
                                logging.debug("func_wrapper: result: %s", result)

                                # Unwrap any partials built up by stacked decorators
                                if callable(result):
                                    return func_wrapper(result, _wq)
                                try:
                                    logging.debug("func_wrapper: putting result on queue")
                                    # TODO: Put (graph,result) tuple here
                                    _wq.put(
                                        {'graph': [(func.__name__, _wf.__name__)], 'result': result})
                                    logging.debug("func_wrapper: done putting queue")
                                except Exception:
                                    with open('error.out', 'w') as errfile:
                                        errfile.write(traceback.format_exc())

                                return None

                            _pfunc = partial(func, *args, **kwargs)
                            proc = multiprocessing.Process(
                                target=func_wrapper, args=(_pfunc, _mq, ))
                            proc.start()
                        else:
                            result = func(*args, **kwargs)
                    else:
                        if event:
                            logging.debug(
                                "Setting event for %s", func.__name__)
                            event.set()
                        return (args, kwargs)
                finally:
                    # Put own cpu back on queue
                    if scheduler and cpu:
                        logging.debug(
                            "Putting CPU: %s back on scheduler queue.", cpu)

                        scheduler.put(['0', cpu, 'N'])

                if self.cache:
                    pass

                logging.debug("PUT GRAPH [%s]: %s", func.__name__, graphs)
                logging.debug(
                    "PUT GRAPH JSON [%s]: %s", func.__name__, json_graph)
                queue.put(
                    {'graph': graphs, 'json': json.loads(json_graph), 'result': result})

                if event:
                    logging.debug(
                        "Setting event for %s", func.__name__)
                    event.set()
            else:
                # Pass in shared memory handles

                if self.shared_memory:
                    kwargs['smm'] = SMM
                    kwargs['sm'] = SharedMemory

                logging.debug(
                    "Calling function %s with: %s", func.__name__, str(args))

                _mq = Queue()

                def func_wrapper(_wf, _wq):
                    logging.debug("func_wrapper: %s", _wf)
                    result = _wf()
                    logging.debug("func_wrapper: result: %s", result)

                    # Unwrap any partials built up by stacked decorators
                    if callable(result):
                        return func_wrapper(result, _wq)
                    try:
                        logging.debug("func_wrapper: putting result on queue")
                        # TODO: Put (graph,result) tuple here
                        _wq.put(
                            {'graph': [(func.__name__, _wf.__name__)], 'result': result})
                        if is_proc:
                            future_queue.put(result)
                        logging.debug("func_wrapper: done putting queue")
                    except Exception:
                        with open('error.out', 'w') as errfile:
                            errfile.write(traceback.format_exc())

                    return None

                pfunc = partial(func, *args, **kwargs)
                pfunc.__name__ = func.__name__

                try:
                    if self.execute:
                        logging.debug("process: execute2: %s", self.execute)
                        proc = multiprocessing.Process(
                            target=func_wrapper, args=(pfunc, _mq,))
                        proc.start()

                        logging.debug(
                            "Executing function %s with timeout %s", func, self.timeout)
                        if not is_proc:
                            proc.join(self.timeout)
                    else:
                        if event:
                            logging.debug(
                                "Setting event for %s", func.__name__)
                            event.set()
                        return (args, kwargs)
                finally:
                    if scheduler and cpu:
                        logging.debug(
                            "Putting CPU: %s back on scheduler queue.", cpu)

                        scheduler.put(('0', cpu, 'Y'))

                logging.debug("process: waiting for result on queue")

                sys.path.append(os.getcwd())

                if not is_proc:
                    response = _mq.get()
                    result = response['result']

                if len(json_graphs) > 0:
                    callgraph = {func.__name__: json_graphs}
                    graph_queue.put(callgraph)
                    self.graph = json.dumps(callgraph)

                logging.debug("process: got result from queue")

                if not is_proc and proc.is_alive():
                    proc.terminate()
                    proc.join()
                    raise ProcessTimeoutException()

            if is_proc:
                # return future
                return True

            future_queue.put(result)

            return result

        # Need to pass in a queue here that the future coroutine can listen on
        kwargs['future_queue'] = self.result_queue
        pfunc = partial(invoke, self.func, *args, **kwargs)

        if hasattr(self.func, '__name__'):
            pfunc.__name__ = self.func.__name__
        else:
            pfunc.__name__ = 'process'

        def get_graph(wait=True):
            if wait:
                return graph_queue.get()
            else:
                @asyncio.coroutine
                def wait_for_graph():

                    logging.debug("wait_for_graph: looping")
                    while True:
                        try:
                            logging.debug("wait_for_graph: checking queue")
                            graph = graph_queue.get_nowait()
                            logging.debug("wait_for_graph: got result")
                            return graph
                        except:
                            yield

                loop = asyncio.get_event_loop()
                task = loop.create_task(wait_for_graph())
                return task

        pfunc.graph = get_graph
        pfunc.future = self.future

        return pfunc  # InvokeProxy(pfunc)



"""
request.py - Module that provides http request tasks
"""
from functools import partial
import requests


def request(function=None,
            timeout=None,
            url=None,
            method='GET',
            sleep=None):
    """

    :param function:
    :param timeout:
    :param url:
    :param method:
    :param sleep:
    :return:
    """
    def decorator(func):
        def wrapper(f_func):

            # Build http request function here, get result
            # call func with result
            def invoke_request(_func, **kwargs):

                def make_request(url, method, data):

                    if method == 'GET':
                        response = requests.get(url=url, params=data)
                        return response.content

                    return None

                response = make_request(url, method, kwargs)
                return _func(response)

            pfunc = partial(invoke_request, f_func)
            pfunc.__name__ = func.__name__
            return pfunc

        return wrapper(func)

    if function is not None:
        return decorator(function)

    return decorator



"""
ssh.py - Run processes and schedulers on remote machines
"""
import logging
import hashlib
import sys
import os
import ast
import re
import codecs
import pickle
import importlib
from functools import partial
from ast import FunctionDef
from uuid import uuid4

import astunparse
import paramiko
from scp import SCPClient
from entangle.process import ProcessMonitor
from entangle.thread import ThreadMonitor


def ssh(function=None, **kwargs):
    """
    Description
    :param function:
    :param kwargs:
    :return:
    """
    def decorator(func, *args, **kwargs):
        """
        Description
        :param func:
        :param args:
        :param kwargs:
        :return:
        """

        def find_func(p_func):
            if isinstance(p_func, partial):
                return find_func(p_func.func)
            return p_func

        hostname = kwargs['host']
        username = kwargs['user']
        sshkey = kwargs['key']
        python = kwargs['python'] if 'python' in kwargs else "python3.8"

        logging.debug("ssh: func: %s", func.func)
        logging.debug("ssh: func source:\n%s", func.source)

        _ast = ast.parse(func.source)

        decorators = _ast.body[0].decorator_list
        ssh_decorator = None

        for decorator in decorators:
            if hasattr(decorator, 'func') and decorator.func.id == 'ssh':
                logging.debug("REMOVE SSH DECORATOR:")
                ssh_decorator = decorator

        if ssh_decorator:
            decorators.remove(ssh_decorator)

        logging.debug("ssh: func new source: %s", astunparse.unparse(_ast))
        logging.debug("ssh args: %s %s",str(args), str(kwargs))

        logging.debug("SSH: Calling function: %s",str(func))

        def wrapper(f_func, *args, **kwargs):
            """
            Description
            :param f:
            :param args:
            :param kwargs:
            :return:
            """
            def remove_ssh_decorator(fsource, username, hostname):
                """
                Description
                :param fsource:
                :param username:
                :param hostname:
                :return:
                """
                _ast = ast.parse(fsource)

                ssh_decorators = []

                for segment in _ast.body:
                    if isinstance(segment, FunctionDef):
                        decorators = segment.decorator_list
                        for decorator in decorators:
                            if hasattr(decorator, 'func') and decorator.func.id == 'ssh':
                                user_keyword = None
                                host_keyword = None
                                for keyword in decorator.keywords:
                                    if keyword.arg == 'user' and keyword.value.value == username:
                                        user_keyword = keyword
                                    if keyword.arg == 'host' and keyword.value.value == hostname:
                                        host_keyword = keyword
                                logging.debug("REMOVE SSH DECORATOR:")

                                if user_keyword and host_keyword:
                                    ssh_decorators += [decorator]

                for ssh_decorator in ssh_decorators:
                    logging.debug("SSH: Removing decorator: %s",ssh_decorator)
                    decorators.remove(ssh_decorator)

                return astunparse.unparse(_ast)

            if 'SOURCE' in os.environ:
                sourcefile = os.environ['SOURCE']
            else:
                sourcefile = sys.argv[0]

            with open(sourcefile) as source:
                logging.debug("SOURCE:%s",source.read())

            logging.debug("SSH: user:%s host:%s key: %s",
                kwargs['user'], kwargs['host'], kwargs['key'])


            def setup_virtualenv(host, user, key, env):
                _ssh = paramiko.SSHClient()
                _ssh.set_missing_host_key_policy(paramiko.AutoAddPolicy())
                _ssh.connect(hostname=hostname, username=username,
                            key_filename=sshkey)
                command = "python3.8 -m venv {}; export LLVM_CONFIG=/usr/bin/llvm-config-10; {}/bin/pip install --upgrade py-entangle".format(
                    env, env)
                _, stdout, _ = _ssh.exec_command(command)
                for line in stdout.read().splitlines():
                    logging.debug("SSH: setup_virtualenv: stdout: %s", line)

            if 'env' in kwargs:
                # Set up virtualenv
                setup_virtualenv(kwargs['host'],kwargs['user'],kwargs['key'],kwargs['env'])
                #python = "{}/bin/python".format(kwargs['env'])
                del kwargs['env']

            del kwargs['user']
            del kwargs['host']
            del kwargs['key']

            logging.debug("Run function: %s(%s,%s)",
                func.__name__, args, kwargs)

            vargs = []

            for arg in args:
                if callable(arg):
                    vargs += [arg()]
                else:
                    vargs += [arg]

            sourceuuid = "sshsource"+hashlib.md5(uuid4().bytes).hexdigest()

            with open(sourcefile) as source:
                _source = source.read()
                _source = re.sub(r"@ssh\(user='{}', host='{}'".format(username,
                                 hostname), "#@ssh(user='{}', host='{}'".format(username,
                                 hostname), _source).strip()

                with open('{}.py'.format(sourceuuid), 'w') as appsource:
                    appsource.write(_source)

            appuuid = "sshapp"+hashlib.md5(uuid4().bytes).hexdigest()

            with open('{}.py'.format(appuuid),'w') as app:
                pargs = codecs.encode(pickle.dumps(vargs), "base64").decode()
                pargs = re.sub(r'\n', "", pargs).strip()
                app.write("import logging\n\n" \
                          "logger=logging.getLogger()\n" \
                          "logging.disabled=False\n" \
                          "logger.disabled=False\n")
                app.write("import pickle, codecs, re\n")
                app.write("from {} import {}\n\n".format(sourceuuid, func.__name__))
                app.write("pargs = '{}'\n".format(pargs))
                app.write("args = pickle.loads(codecs.decode(pargs.encode(), \"base64\"))\n")
                app.write("print(\"ARGS: {}\".format(args))\n")
                app.write("result = {}(*args)()\n".format(func.__name__))
                app.write("print(\"RESULT:\", result)\n")
                app.write(
                    "resultp = codecs.encode(pickle.dumps(result), \"base64\").decode()\n")
                app.write("print('===BEGIN===')\n")
                app.write("print(resultp)")

            p_func = partial(f_func, *args, **kwargs)

            p_func.__name__ = f_func.__name__

            logging.debug("args: %s   kwargs: %s",args,kwargs)

            def ssh_function(remotefunc, username, hostname, sshkey, appuuid, sourceuuid):
                """
                Description
                :param remotefunc:
                :param username:
                :param hostname:
                :param sshkey:
                :param appuuid:
                :param sourceuuid:
                :return:
                """
                files = [appuuid+".py", sourceuuid+".py"]
                logging.debug(
                    "SCP files: %s to %s@%s:%s", files, username, hostname, '/tmp')
                _ssh = paramiko.SSHClient()
                _ssh.set_missing_host_key_policy(paramiko.AutoAddPolicy())
                _ssh.connect(hostname=hostname, username=username,
                            key_filename=sshkey)
                logging.debug("SSH FUNCTION: %s",remotefunc)
                scp = SCPClient(_ssh.get_transport())

                try:
                    scp.put(files,
                            remote_path='/tmp')
                finally:
                    pass
                    #[os.remove(file) for file in files]

                command = "export SOURCE={}.py; cd /tmp; {} /tmp/{}.py".format(sourceuuid,
                    python, appuuid)

                result = None

                logging.debug("SSH: executing %s %s@%s",command, username, hostname)
                _, stdout, _ = _ssh.exec_command(command)

                logging.debug("SSH: CWD %s",os.getcwd())
                logging.debug("SSH: importing module %s",sourceuuid)
                sys.path.append(os.getcwd())
                try:
                    importlib.import_module(sourceuuid)
                except Exception:
                    pass

                result_next = False
                resultlines = []
                for line in stdout.read().splitlines():
                    logging.debug("SSH: command stdout: %s",line)
                    if result_next:
                        if len(line.strip()) > 0:
                            resultlines += [line]
                            logging.debug("SSH: got result line: %s",line)
                    if line == b"===BEGIN===":
                        result_next = True

                logging.debug("Unpickle: %s",b"".join(resultlines))

                result = pickle.loads(
                    codecs.decode(b"".join(resultlines), "base64"))

                _ssh.exec_command("rm {}".format(" ".join(["/tmp/"+file for file in files])))
                _ssh.close()

                return result

            ssh_p = partial(ssh_function, p_func, username,
                            hostname, sshkey, appuuid, sourceuuid)
            ssh_p.__name__ = p_func.__name__

            result = ProcessMonitor(ssh_p, timeout=None,
                                    wait=None,
                                    cache=False,
                                    shared_memory=False,
                                    sleep=0)

            if callable(result):
                _result = result()
            else:
                _result = result

            logging.debug("SSH RESULT: %s",_result)

            # files = [appuuid+".py", sourceuuid+".py"]
            # [os.remove(file) for file in files]
            return _result

        p_func = partial(wrapper, func, **kwargs)

        if isinstance(func, (ProcessMonitor, ThreadMonitor)):
            p_func.__name__ = func.func.__name__
        else:
            p_func.__name__ = func.__name__

        return p_func

    if function is not None:
        logging.debug("ssh: function source: %s", function.source)
        return decorator(function, **kwargs)

    return partial(decorator, **kwargs)



# pylint: disable=locally-disabled, multiple-statements, no-value-for-parameter, invalid-name, too-many-function-args, unused-import, missing-function-docstring
"""
TBD
"""
from entangle.logging.file import logging
from entangle.process import process
from entangle.http import request
from entangle.workflow import workflow
from entangle.scheduler import scheduler

scheduler_config = {'cpus': 2,
                    'impl': 'entangle.scheduler.DefaultScheduler'}


@scheduler(**scheduler_config)
@process
def two():
    return 2


@scheduler(**scheduler_config)
@process
def three():
    return 3


@scheduler(**scheduler_config)
@process
def add(a, b):
    print("add: {} {}".format(a,b))
    v = int(a) + int(b)
    print("ADD: *"+str(v)+"*")
    return v


@scheduler(**scheduler_config)
@workflow
def workflow2():
    return add(
        three(),
        two()
    )


result = workflow2()

print(result())



#!/usr/bin/env python3

import os
import glob
import shutil
from setuptools import setup
from setuptools import Command
import distutils.cmd
import distutils.log
import setuptools
import subprocess

# get key package details from py_pkg/__version__.py
about = {}  # type: ignore
here = os.path.abspath(os.path.dirname(__file__))
with open(os.path.join(here, 'entangle', '__version__.py')) as f:
    exec(f.read(), about)

# load the README file and use it as the long_description for PyPI
with open('README.md', 'r') as f:
    readme = f.read()


class PyTestCommand(distutils.cmd.Command):
  """A custom command to run Pylint on all Python source files."""

  description = 'Run pytest tests'
  user_options = [
      
  ]

  def initialize_options(self):
    """Set default values for options."""
    # Each user option must be listed here with their default value.
    pass

  def finalize_options(self):
    """Post-process options."""
    pass

  def run(self):
    """Run command."""
    command = [
        'pytest --full-trace --verbose --color=yes --disable-pytest-warnings --no-summary --pyargs entangle.tests']

    self.announce(
        'Running command: %s' % str(command),
        level=distutils.log.INFO)
    subprocess.Popen(command, shell=True)

class PylintCommand(distutils.cmd.Command):
  """A custom command to run Pylint on all Python source files."""

  description = 'run Pylint on Python source files'
  user_options = [
      # The format is (long option, short option, description).
      ('pylint-rcfile=', None, 'path to Pylint config file'),
  ]

  def initialize_options(self):
    """Set default values for options."""
    # Each user option must be listed here with their default value.
    self.pylint_rcfile = ''

  def finalize_options(self):
    """Post-process options."""
    if self.pylint_rcfile:
      assert os.path.exists(self.pylint_rcfile), (
          'Pylint config file %s does not exist.' % self.pylint_rcfile)

  def run(self):
    """Run command."""
    command = ['pylint entangle']
    #if self.pylint_rcfile:
    #  command.append('--rcfile=%s' % self.pylint_rcfile)
    #command.append('entangle')
    self.announce(
        'Running command: %s' % str(command),
        level=distutils.log.INFO)
    subprocess.Popen(command, shell=True)

class CleanCommand(Command):
    """Custom clean command to tidy up the project root."""
    CLEAN_FILES = './build ./dist ./__pycache__ **/*/__pycache__ ./*.pyc ./ssh*py ./*.tgz ./entangle.log ./.pytest_cache ./*.egg-info'.split(' ')

    user_options = []

    def initialize_options(self):
        pass

    def finalize_options(self):
        pass

    def run(self):
        global here

        for path_spec in self.CLEAN_FILES:
            # Make paths absolute and relative to this path
            abs_paths = glob.glob(os.path.normpath(
                os.path.join(here, path_spec)))
            for path in [str(p) for p in abs_paths]:
                if not path.startswith(here):
                    # Die if path in CLEAN_FILES is absolute + outside this directory
                    raise ValueError(
                        "%s is not a path inside %s" % (path, here))
                print('removing %s' % os.path.relpath(path))
                try:
                    shutil.rmtree(path)
                except:
                    os.remove(path)


# package configuration - for reference see:
# https://setuptools.readthedocs.io/en/latest/setuptools.html#id9
setup(
    name=about['__title__'],
    description=about['__description__'],
    long_description=readme,
    long_description_content_type='text/markdown',
    version=about['__version__'],
    author=about['__author__'],
    author_email=about['__author_email__'],
    url=about['__url__'],
    packages=['entangle', 'entangle.logging', 'entangle.examples', 'entangle.tests'],
    include_package_data=True,
    python_requires="==3.8.*",
    install_requires=[
        'scp==0.13.3',
        'astunparse==1.6.3',
        'async-timeout==3.0.1',
        'cachetools==4.2.1',
        'certifi==2020.12.5',
        'cffi==1.14.3',
        'cryptography==3.2.1',
        'dill==0.3.3',
        'docker==5.0.0',
        'filelock==3.0.12',
        'gast==0.4.0',
        'globus-sdk==2.0.1',
        'numpy==1.19.2',
        #'numba==0.53.1',
        'paramiko==2.7.2',
        'pip',
        'psutil==5.8.0',
        'py-cpuinfo==8.0.0',
        'pytest==6.2.4',
        'requests==2.24.0',
        'scipy==1.6.2',
        'setuptools',
        'six==1.15.0',
        'typeguard==2.12.0',
        'pylint==2.8.2',
        'astunparse==1.6.3'
    ],
    license=about['__license__'],
    zip_safe=False,
    cmdclass={
        'clean': CleanCommand,
        'pylint': PylintCommand,
        'test': PyTestCommand
    },
    entry_points={

    },
    classifiers=[
        'Development Status :: 2 - Pre-Alpha',
        'Intended Audience :: Developers',
        'Programming Language :: Python :: 3.8',
    ],
    keywords='parallel processing, devops, dataflow, supercomputing, workflows'
)



# pylint: disable=locally-disabled, multiple-statements, no-value-for-parameter, no-member, invalid-name, too-many-function-args, unused-import, missing-function-docstring
"""
TBD
"""
from timeit import default_timer as timer
import numpy as np
from numba import vectorize
from entangle.logging.file import logging
from entangle.process import process
from entangle.scheduler import scheduler


scheduler_config = {'cpus': 3,
                    'impl': 'entangle.scheduler.DefaultScheduler'}


#@scheduler(**scheduler_config)
@process
def dovectors1():
    # pylint: disable=locally-disabled, unused-variable

    @vectorize(['float32(float32, float32)'], target='cuda')
    def dopow(a, b):
        return a ** b

    vec_size = 100

    a = b = np.array(np.random.sample(vec_size), dtype=np.float32)
    c = np.zeros(vec_size, dtype=np.float32)

    start = timer()
    matrix = dopow(a, b)
    duration = timer() - start
    return duration


#@scheduler(**scheduler_config)
@process
def dovectors2():
    # pylint: disable=locally-disabled, unused-variable

    @vectorize(['float32(float32, float32)'], target='cuda')
    def dopow(a, b):
        return a ** b

    vec_size = 100000000

    a = b = np.array(np.random.sample(vec_size), dtype=np.float32)
    c = np.zeros(vec_size, dtype=np.float32)

    start = timer()
    matrix = dopow(a, b)
    duration = timer() - start
    return duration


#@scheduler(**scheduler_config)
@process
def durations(*args):

    return args


if __name__ == '__main__':
    dp = durations(
        dovectors1(),
        dovectors2()
    )

    print(dp())



"""
workflow.py - Module that provides workflow decorator
"""
import logging


def workflow(func, **kwargs):
    """
    Desc
    :param func:
    :param kwargs:
    :return:
    """
    def inner(*args, **kwargs):
        """
        Desc
        :param args:
        :param kwargs:
        :return:
        """
        logging.debug("workflow args: %s %s",str(args), str(kwargs))

        if 'cpu' in kwargs:
            del kwargs['cpu']

        _workflow = func(*args, **kwargs)
        # added QoS implementation here
        logging.debug("Inside the workflow")
        return _workflow

    return inner



# pylint: disable=locally-disabled, multiple-statements, unexpected-keyword-arg, no-value-for-parameter, invalid-name, too-many-function-args, unused-import, missing-function-docstring
"""
TBD
"""
import json
from entangle.logging.debug import logging
from entangle.scheduler import scheduler
from entangle.workflow import workflow
from entangle.http import request
from entangle.thread import thread

scheduler_config = {'cpus': 10, 'impl': 'entangle.scheduler.DefaultScheduler'}


@scheduler(**scheduler_config)
@thread
@request(url='https://datausa.io/api/data', method='GET')
def mydata(data):
    data = json.loads(data)
    print('*********************My function got the data! ', data)
    return int(data['data'][0]['Year'])


@scheduler(**scheduler_config)
@thread
def two():
    return 2


@scheduler(**scheduler_config)
@thread
def add(a, b):
    v = int(a) + int(b)
    print("ADD: *"+str(v)+"*")
    return v


@scheduler(**scheduler_config)
@workflow
def workflow1():
    return add(
        mydata(drilldowns='Nation', measures='Population'),
        two()
    )


@scheduler(**scheduler_config)
@workflow
def workflow2(value):
    return add(
        value(),
        two()
    )


result = workflow2(workflow1)

print(result())
